<?php
/**
 * SPDX-FileCopyrightText: 2024 Sascha Nitsch (grumpydeveloper) https://contentnation.net/@grumpydevelop
 * SPDX-License-Identifier: GPL-3.0-or-later
 *
 * @author Yannis Vogel (vogeldevelopment)
 **/

namespace Federator\Api\FedUsers;

/**
 * handle activitypub followers requests
 */
class Followers implements \Federator\Api\FedUsers\FedUsersInterface
{
    /**
     * main instance
     *
     * @var \Federator\Api $main
     */
    private $main;

    /**
     * constructor
     * @param \Federator\Api $main main instance
     */
    public function __construct($main)
    {
        $this->main = $main;
    }

    /**
     * handle get call
     *
     * @param string|null $_user user to fetch followers for
     * @return string|false response
     */
    public function get($_user)
    {
        if (!isset($_user)) {
            return false;
        }
        $dbh = $this->main->getDatabase();
        $cache = $this->main->getCache();
        $connector = $this->main->getConnector();

        // get user
        $user = \Federator\DIO\User::getUserByName(
            $dbh,
            $_user,
            $connector,
            $cache
        );
        if ($user->id === null) {
            return false;
        }

        $followers = new \Federator\Data\ActivityPub\Common\Followers();
        $followerItems = \Federator\DIO\Followers::getFollowersByUser($dbh, $user->id, $connector, $cache);

        $config = $this->main->getConfig();
        $protocol = $config['generic']['protocol'];
        $domain = $config['generic']['externaldomain'];
        $baseUrl = $protocol . '://' . $domain . '/' . $_user . '/followers';

        $pageSize = 10;
        $page = $this->main->extractFromURI('page', '');
        $id = $baseUrl;
        $items = [];
        $totalItems = count($followerItems);

        if ($page !== '') {
            $pageNum = max(0, (int) $page);
            $offset = (int)($pageNum * $pageSize);
            $pagedItems = array_slice($followerItems, $offset, $pageSize);

            foreach ($pagedItems as $follower) {
                $items[] = $follower->actorURL;
            }
            $followers->setItems($items);
            $id .= '?page=' . urlencode($page);
        }

        $followers->setID($id);
        $followers->setPartOf($baseUrl);
        $followers->setTotalItems($totalItems);

        // Pagination navigation
        $lastPage = max(0, ceil($totalItems / $pageSize) - 1);

        if ($page === '' || $followers->getCount() == 0) {
            $followers->setFirst($baseUrl . '?page=0');
            $followers->setLast($baseUrl . '?page=' . $lastPage);
        }
        if ($page !== '') {
            $pageNum = max(0, (int) $page);
            if ($pageNum < $lastPage) {
                $followers->setNext($baseUrl . '?page=' . ($pageNum + 1));
            }
            if ($pageNum > 0) {
                $followers->setPrev($baseUrl . '?page=' . ($pageNum - 1));
            }
            $followers->setFirst($baseUrl . '?page=0');
            $followers->setLast($baseUrl . '?page=' . $lastPage);
        } else {
            $followers->setType('OrderedCollection');
        }
        $obj = $followers->toObject();

        return json_encode($obj, JSON_UNESCAPED_SLASHES | JSON_PRETTY_PRINT);
    }

    /**
     * handle post call
     *
     * @param string|null $_user user to add data to outbox @unused-param
     * @return string|false response
     */
    public function post($_user)
    {
        return false;
    }
}
